from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.constants import Workflow
from cg.store.models import Application, ApplicationLimitations


def filter_application_limitations_by_tag(
    application_limitations: Query, tag: str, **kwargs
) -> Query:
    """Return application limitations by tag."""
    return application_limitations.filter(Application.tag == tag)


def filter_application_limitations_by_workflow(
    application_limitations: Query, workflow: Workflow, **kwargs
) -> Query:
    """Return application limitations by workflow."""
    return application_limitations.filter(ApplicationLimitations.pipeline == workflow)


def apply_application_limitations_filter(
    filter_functions: list[Callable],
    application_limitations: Query,
    tag: str = None,
    workflow: Workflow = None,
) -> Query:
    """Apply filtering functions to the application limitations queries and return filtered results."""
    for filter_function in filter_functions:
        application_limitations: Query = filter_function(
            application_limitations=application_limitations,
            tag=tag,
            workflow=workflow,
        )
    return application_limitations


class ApplicationLimitationsFilter(Enum):
    """Define ApplicationLimitations filter functions."""

    BY_TAG = filter_application_limitations_by_tag
    BY_WORKFLOW = filter_application_limitations_by_workflow



"""Tests for the deliver base command"""

import logging

from click.testing import CliRunner

from cg.cli.deliver.base import deliver as deliver_cmd
from cg.constants import EXIT_SUCCESS
from cg.meta.deliver_ticket import DeliverTicketAPI
from cg.models.cg_config import CGConfig


def test_run_base_help(cli_runner: CliRunner):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner

    # WHEN running cg deliver help
    result = cli_runner.invoke(deliver_cmd, ["--help"])

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "Deliver files with CG" in result.output


def test_run_deliver_analysis_help(cli_runner: CliRunner, base_context: CGConfig):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner
    # GIVEN a context with store and housekeeper information

    # WHEN running cg deliver help
    result = cli_runner.invoke(deliver_cmd, ["analysis", "--help"], obj=base_context)

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "Deliver analysis files to customer inbox" in result.output


def test_run_deliver_ticket_help(cli_runner: CliRunner, base_context: CGConfig):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner
    # GIVEN a context with store and housekeeper information

    # WHEN running cg deliver help
    result = cli_runner.invoke(deliver_cmd, ["ticket", "--help"], obj=base_context)

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "Will first collect hard links" in result.output


def test_run_deliver_delivered_ticket(
    cli_runner: CliRunner, cg_context: CGConfig, mocker, caplog, ticket_id
):
    """Test for when files are already delivered to customer inbox the HPC"""
    caplog.set_level(logging.INFO)

    # GIVEN a cli runner

    # GIVEN uploading data to the delivery server is not needed
    mocker.patch.object(DeliverTicketAPI, "check_if_upload_is_needed")
    DeliverTicketAPI.check_if_upload_is_needed.return_value = False

    # WHEN running cg deliver ticket
    result = cli_runner.invoke(
        deliver_cmd,
        ["ticket", "--dry-run", "--ticket", ticket_id, "--delivery-type", "fastq"],
        obj=cg_context,
    )

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS

    # THEN assert that files are already delivered to the customer inbox on the HPC
    assert "Files already delivered to customer inbox on the HPC" in caplog.text


def test_deliver_ticket_with_force_all_flag(
    cli_runner: CliRunner, cg_context: CGConfig, mocker, caplog, ticket_id
):
    """Test that when the --force-all flag is used,
    the files are delivered to the customer inbox on the HPC"""
    caplog.set_level(logging.INFO)

    # GIVEN a cli runner

    # GIVEN uploading data to the delivery server is not needed
    mocker.patch.object(DeliverTicketAPI, "check_if_upload_is_needed")
    DeliverTicketAPI.check_if_upload_is_needed.return_value = False

    # WHEN running cg deliver ticket with --force-all flag
    cli_runner.invoke(
        deliver_cmd,
        ["ticket", "--dry-run", "--ticket", ticket_id, "--delivery-type", "fastq", "--force-all"],
        obj=cg_context,
    )

    # THEN assert that the text is not present in the log
    assert "Files already delivered to customer inbox on the HPC" not in caplog.text
    assert "Delivering files to customer inbox on the HPC" in caplog.text


def test_run_deliver_ticket(cli_runner: CliRunner, cg_context: CGConfig, mocker, caplog, ticket_id):
    """Test for delivering tu customer inbox"""
    caplog.set_level(logging.INFO)

    # GIVEN a cli runner

    # GIVEN uploading data to the delivery server is needed
    mocker.patch.object(DeliverTicketAPI, "check_if_upload_is_needed")
    DeliverTicketAPI.check_if_upload_is_needed.return_value = True

    # GIVEN data needs to be concatenated
    mocker.patch.object(DeliverTicketAPI, "check_if_concatenation_is_needed")
    DeliverTicketAPI.check_if_concatenation_is_needed.return_value = True

    # WHEN running cg deliver ticket
    cli_runner.invoke(
        deliver_cmd,
        ["ticket", "--dry-run", "--ticket", ticket_id, "--delivery-type", "fastq"],
        obj=cg_context,
    )

    # THEN assert that files are delivered
    assert "Delivering files to customer inbox on the HPC" in caplog.text



"""The JanusAPIClient."""

from http import HTTPStatus

import requests
from requests import Response

from cg.clients.janus.dto.create_qc_metrics_request import CreateQCMetricsRequest
from cg.clients.janus.exceptions import JanusClientError, JanusServerError


class JanusAPIClient:
    def __init__(self, config: dict[str]):
        self.host = config["janus"]["host"]

    def qc_metrics(self, collect_qc_request: CreateQCMetricsRequest) -> dict | None:
        endpoint: str = f"{self.host}/collect_qc_metrics"
        post_request_data: str = collect_qc_request.model_dump_json()
        response = requests.post(endpoint, data=post_request_data)
        if response.status_code == HTTPStatus.OK:
            return response.json()
        self._handle_errors(response)

    @staticmethod
    def _handle_errors(response: Response):
        if 400 <= response.status_code < 500:
            raise JanusClientError(f"Client error: {response.status_code}")
        elif 500 <= response.status_code:
            raise JanusServerError(f"Server error: {response.status_code}")



import logging

import click

from cg.cli.workflow.commands import ARGUMENT_CASE_ID
from cg.cli.workflow.fastq.fastq_service import FastqService
from cg.constants.constants import DRY_RUN, Workflow
from cg.meta.workflow.analysis import AnalysisAPI
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def fastq(context: click.Context):
    """Function for storing fastq-cases"""
    AnalysisAPI.get_help(context)


@fastq.command("store")
@DRY_RUN
@ARGUMENT_CASE_ID
@click.pass_context
def store_fastq_analysis(context: click.Context, case_id: str, dry_run: bool = False):
    LOG.info(f"Creating an analysis for case {case_id}")

    if dry_run:
        return

    fastq_service = FastqService(
        store=context.obj.status_db,
        trailblazer_api=context.obj.trailblazer_api,
    )
    fastq_service.store_analysis(case_id)


@fastq.command("store-available")
@DRY_RUN
@click.pass_context
def store_available_fastq_analysis(context: click.Context, dry_run: bool = False):
    """Creates an analysis object in status-db for all fastq cases to be delivered"""
    status_db: Store = context.obj.status_db
    for case in status_db.cases_to_analyze(workflow=Workflow.FASTQ, threshold=False):
        context.invoke(store_fastq_analysis, case_id=case.internal_id, dry_run=dry_run)



import logging
from pathlib import Path

from housekeeper.store.models import File
from pydantic import BaseModel, Field

from cg.meta.archive.ddn.constants import OSTYPE, ROOT_TO_TRIM, JobStatus
from cg.meta.archive.models import FileTransferData
from cg.store.models import Sample

LOG = logging.getLogger(__name__)


def get_request_log(body: dict):
    return "Sending request with body: \n" + f"{body}"


class MiriaObject(FileTransferData):
    """Model for representing a singular object transfer."""

    destination: str
    source: str

    @classmethod
    def create_from_file_and_sample(
        cls, file: File, sample: Sample, is_archiving: bool = True
    ) -> "MiriaObject":
        """Instantiates the class from a File and Sample object."""
        if is_archiving:
            return cls(destination=sample.internal_id, source=file.full_path)
        return cls(
            destination=Path(file.full_path).parent.as_posix(),
            source=Path(sample.internal_id, Path(file.path).name).as_posix(),
        )

    def trim_path(self, attribute_to_trim: str):
        """Trims the given attribute (source or destination) from its root directory."""
        setattr(
            self,
            attribute_to_trim,
            f"/{Path(getattr(self, attribute_to_trim)).relative_to(ROOT_TO_TRIM)}",
        )

    def add_repositories(self, source_prefix: str, destination_prefix: str):
        """Prepends the given repositories to the source and destination paths."""
        self.source: str = source_prefix + self.source
        self.destination: str = destination_prefix + self.destination


class TransferPayload(BaseModel):
    """Model for representing a Dataflow transfer task."""

    files_to_transfer: list[MiriaObject] = Field(..., serialization_alias="pathInfo")
    osType: str = OSTYPE
    createFolder: bool = True
    settings: list[dict] = []
    metadataList: list[dict] = []

    def trim_paths(self, attribute_to_trim: str):
        """Trims the source path from its root directory for all objects in the transfer."""
        for miria_file in self.files_to_transfer:
            miria_file.trim_path(attribute_to_trim=attribute_to_trim)

    def add_repositories(self, source_prefix: str, destination_prefix: str):
        """Prepends the given repositories to the source and destination paths all objects in the
        transfer."""
        for miria_file in self.files_to_transfer:
            miria_file.add_repositories(
                source_prefix=source_prefix, destination_prefix=destination_prefix
            )


class ArchivalResponse(BaseModel):
    """Model representing the response fields of an archive request to the Dataflow
    API."""

    job_id: int = Field(alias="jobId")


class RetrievalResponse(BaseModel):
    """Model representing the response fields of a retrieval reqeust to the Dataflow
    API."""

    job_id: int = Field(alias="jobId")


class AuthPayload(BaseModel):
    """Model representing the payload for an Authentication request."""

    dbName: str
    name: str
    password: str
    superUser: bool = False


class RefreshPayload(BaseModel):
    """Model representing the payload for Auth-token refresh request."""

    refresh: str


class AuthToken(BaseModel):
    """Model representing the response fields from an access request to the Dataflow API."""

    access: str
    expire: int
    refresh: str | None = None


class GetJobStatusResponse(BaseModel):
    """Model representing the response fields from a get_job_status post."""

    job_id: int = Field(alias="id")
    status: JobStatus


class DeleteFileResponse(BaseModel):
    message: str


class DeleteFilePayload(BaseModel):
    global_path: str



"""Tests CLI common methods to create the report_deliver for NF analyses."""

import logging
import pytest
from pathlib import Path

from click.testing import CliRunner
from _pytest.logging import LogCaptureFixture

from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.cli.workflow.rnafusion.base import report_deliver
from cg.models.cg_config import CGConfig
from cg.constants import EXIT_SUCCESS
from cg.io.controller import ReadFile
from cg.constants.constants import FileFormat

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "context",
    ["rnafusion_context", "taxprofiler_context"],
)
def test_report_deliver_without_options(cli_runner: CliRunner, context: CGConfig, request):
    """Test report-deliver for workflow without options."""
    # GIVEN a NextFlow analysis context
    context = request.getfixturevalue(context)

    # WHEN dry running without anything specified
    result = cli_runner.invoke(report_deliver, obj=context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN command log should inform about missing arguments
    assert "Missing argument" in result.output


@pytest.mark.parametrize(
    "context",
    ["rnafusion_context", "taxprofiler_context"],
)
def test_report_deliver_with_missing_case(
    cli_runner: CliRunner,
    context: CGConfig,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
    request,
):
    """Test report-deliver command for workflow with invalid case to start with."""
    caplog.set_level(logging.WARNING)
    context = request.getfixturevalue(context)

    # GIVEN case_id not in database
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id_does_not_exist)

    # WHEN generating deliverables file
    result = cli_runner.invoke(report_deliver, [case_id_does_not_exist], obj=context)

    # THEN command should NOT succeed
    assert result.exit_code != EXIT_SUCCESS

    # THEN ERROR log should be printed containing the invalid case id
    assert case_id_does_not_exist in caplog.text
    assert "could not be found" in caplog.text


@pytest.mark.parametrize(
    "context",
    ["rnafusion_context", "taxprofiler_context"],
)
def test_report_deliver_without_samples(
    cli_runner: CliRunner,
    context: CGConfig,
    caplog: LogCaptureFixture,
    no_sample_case_id: str,
    request,
):
    """Test report-deliver command for workflow with case id and no samples."""
    caplog.set_level(logging.ERROR)
    context = request.getfixturevalue(context)

    # GIVEN case-id
    case_id: str = no_sample_case_id

    # WHEN generating deliverables with dry specified
    result = cli_runner.invoke(report_deliver, [case_id, "--dry-run"], obj=context)

    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN warning should be printed that no samples exist in case
    assert "no samples" in caplog.text


@pytest.mark.parametrize(
    ("context", "case_id", "deliverables_file_path", "mock_analysis_finish"),
    [
        (
            "taxprofiler_context",
            "taxprofiler_case_id",
            "taxprofiler_deliverables_file_path",
            "taxprofiler_mock_analysis_finish",
        ),
        (
            "rnafusion_context",
            "rnafusion_case_id",
            "rnafusion_deliverables_file_path",
            "rnafusion_mock_analysis_finish",
        ),
    ],
)
def test_report_deliver_successful(
    cli_runner: CliRunner,
    context: CGConfig,
    case_id: str,
    deliverables_file_path: Path,
    deliverables_template_content: list[dict],
    mock_analysis_finish,
    caplog: LogCaptureFixture,
    mocker,
    request,
):
    """Test that deliverable files is properly created on a valid and successful run for workflow."""
    caplog.set_level(logging.INFO)

    # GIVEN each fixture is being initialised
    context: CGConfig = request.getfixturevalue(context)
    deliverables_file_path: dict = request.getfixturevalue(deliverables_file_path)
    case_id: str = request.getfixturevalue(case_id)
    request.getfixturevalue(mock_analysis_finish)

    # GIVEN a successful run

    # GIVEN a mocked deliverables template
    mocker.patch.object(
        NfAnalysisAPI,
        "get_deliverables_template_content",
        return_value=deliverables_template_content,
    )

    # WHEN running the command
    result = cli_runner.invoke(report_deliver, [case_id], obj=context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN deliverables file should be written
    assert deliverables_file_path.is_file()

    # THEN deliverables content should match the expected values
    deliverables_content: str = ReadFile.get_content_from_file(
        file_format=FileFormat.TXT, file_path=deliverables_file_path, read_to_string=True
    )
    for field in deliverables_template_content[0].keys():
        assert field in deliverables_content

    # THEN assess that missing fields are written
    assert "path_index: null" in deliverables_content



"""Model MIP sample info"""

import datetime

from pydantic.v1 import BaseModel, Field, validator


class MipBaseSampleInfo(BaseModel):
    """This model is used when validating the mip sample info file"""

    family_id_: str | None = Field(None, alias="family_id")
    case_id: str | None
    human_genome_build_: dict = Field(..., alias="human_genome_build")
    genome_build: str = None
    program_: dict = Field(None, alias="program")
    recipe_: dict = Field(None, alias="recipe")
    rank_model_version: str = None
    sv_rank_model_version: str = None
    analysis_run_status_: str = Field(..., alias="analysisrunstatus")
    date: datetime.datetime = Field(..., alias="analysis_date")
    is_finished: bool = False
    mip_version: str = None

    @validator("case_id", always=True, pre=True)
    def set_case_id(cls, value, values: dict) -> str:
        """Set case_id. Family_id is used for older versions of MIP analysis"""
        return value or values.get("family_id_")

    @validator("genome_build", always=True, pre=True)
    def set_genome_build(cls, _, values: dict) -> str:
        """Set genome_build by combining source i.e. "hg"|"grch" and version"""
        raw_genome_build: dict = values.get("human_genome_build_")
        source = raw_genome_build.get("source")
        version = raw_genome_build.get("version")
        return f"{source}{version}"

    @validator("is_finished", always=True)
    def set_is_finished(cls, _, values: dict) -> bool:
        """Set is_finished from analysisrunstatus_"""
        return values.get("analysis_run_status_") == "finished"

    @validator("rank_model_version", always=True)
    def set_rank_model_version(cls, _, values: dict) -> str:
        """Set rank_model_version from genmod snv/indel analysis"""
        if "recipe_" in values:
            if "genmod" in values["recipe_"]:
                return values["recipe_"]["genmod"]["rank_model"]["version"]
        else:
            if "genmod" in values["program_"]:
                return values["program_"]["genmod"]["rank_model"]["version"]

    @validator("sv_rank_model_version", always=True)
    def set_sv_rank_model_version(cls, _, values: dict) -> str:
        """Set rank_model_version from genmod SV analysis"""
        if "recipe_" in values:
            if "genmod" in values["recipe_"]:
                return values["recipe_"]["sv_genmod"]["sv_rank_model"]["version"]
        else:
            if "genmod" in values["program_"]:
                return values["program_"]["sv_genmod"]["sv_rank_model"]["version"]



from pathlib import Path

from cg.apps.downsample.downsample import DownsampleAPI
from cg.apps.downsample.utils import store_downsampled_sample_bundle
from cg.models.downsample.downsample_data import DownsampleData


def test_add_fastq_files_to_housekeeper(
    downsample_api: DownsampleAPI, downsample_data: DownsampleData, tmp_path, downsample_dir: Path
):
    """Test to add downsampled fastq files to housekeeper."""

    # GIVEN a downsample api and downsampled fastq files
    downsampled_sample_dir = Path(
        downsample_dir, f"{downsample_data.downsampled_sample.internal_id}"
    )
    downsampled_sample_dir.mkdir()
    Path(
        downsampled_sample_dir,
        f"{downsample_data.downsampled_sample.internal_id}.fastq.gz",
    ).touch()
    assert downsample_data.fastq_file_output_directory.exists()
    # WHEN adding fastq files to housekeeper
    store_downsampled_sample_bundle(
        housekeeper_api=downsample_data.housekeeper_api,
        sample_id=downsample_data.downsampled_sample.internal_id,
        fastq_file_output_directory=str(downsampled_sample_dir),
    )

    # THEN fastq files are added to the downsampled bundle
    assert downsample_api.housekeeper_api.get_latest_bundle_version(
        downsample_data.downsampled_sample.internal_id
    )

    assert downsample_api.housekeeper_api.get_files(
        downsample_data.downsampled_sample.internal_id, tags=["fastq"]
    )



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_fastq_delivery_path
from cg.store.models import Case


class MicrosaltMwrMessage(DeliveryMessage):
    def create_message(self, case: Case) -> str:
        delivery_path: str = get_fastq_delivery_path(case)
        return (
            f"Hello,\n\n "
            f"The analysis is now complete and the fastq files are being uploaded to:\n\n"
            f"{delivery_path} \n\n"
            "The QC and Typing reports can be found attached."
        )



"""Constants specific for Tar extraction"""

from cg.utils.enums import ListEnum


class FlowCellExtractionParameters(ListEnum):
    EXTRACT_FILE: list = [
        "-xf",
    ]
    EXCLUDE_FILES: list = [
        "--exclude=RTAComplete.txt",
        "--exclude=demuxstarted.txt",
        "--exclude=Thumbnail_Images",
    ]
    CHANGE_TO_DIR: list = [
        "-C",
    ]



"""Tests for the metrics-deliver CLI command."""

import logging
from pathlib import Path

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.taxprofiler.base import metrics_deliver
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.io.yaml import read_yaml


def test_metrics_deliver(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    taxprofiler_mock_analysis_finish,
    caplog: LogCaptureFixture,
    taxprofiler_case_id: str,
    taxprofiler_metrics_deliverables_path: Path,
):
    """Test command with a case id and a finished analysis which should execute successfully."""
    caplog.set_level(logging.INFO)

    # GIVEN case id
    result = cli_runner.invoke(metrics_deliver, [taxprofiler_case_id], obj=taxprofiler_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN metrics deliverable file should be generated
    assert taxprofiler_metrics_deliverables_path.is_file()

    # WHEN reading metrics_deliverable_content as yaml
    metrics_deliverables_content = read_yaml(file_path=taxprofiler_metrics_deliverables_path)

    assert isinstance(metrics_deliverables_content, dict)

    # THEN the metrics_deliverables_content contains the key "metrics"
    assert "metrics" in metrics_deliverables_content.keys()

    # THEN taxprofiler and case_id should be found in command string
    assert "Writing metrics deliverables file to" in caplog.text



import logging

from cg.constants.scout import BALSAMIC_UMI_CASE_TAGS
from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class BalsamicUmiReportAPI(BalsamicReportAPI):
    """API to create Balsamic UMI delivery reports."""

    def __init__(self, config: CGConfig, analysis_api: BalsamicUmiAnalysisAPI):
        super().__init__(config=config, analysis_api=analysis_api)
        self.analysis_api: BalsamicUmiAnalysisAPI = analysis_api

    def get_upload_case_tags(self) -> dict:
        """Return Balsamic UMI upload case tags."""
        return BALSAMIC_UMI_CASE_TAGS
